# This file is part of Level-Set-Based-Camera-Pose-Estimation.
#
# Author: Matthieu Zins (matthieu.zins@inria.fr)
#
# Level-Set-Based-Camera-Pose-Estimation is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# Level-Set-Based-Camera-Pose-Estimation is distributed in the hope that
# it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Level-Set-Based-Camera-Pose-Estimation.
# If not, see <http://www.gnu.org/licenses/>.


import argparse
import glob
import json
import logging
import os
import time

import cv2
import numpy as np

import torch
from torchvision import transforms
import pytorch_lightning as pl
from pytorch_lightning import loggers as pl_loggers
from pytorch_lightning.callbacks import ModelCheckpoint

from config.config import _cfg
from dataset.dataset import EllipsesDataset
from dataset.scene_loader import Scene_loader
from ellipse_prediction.model import EllipsePredictor
from ellipse_prediction.model_uncertainty import EllipsePredictor as EllipsePredictor_uncertainty



def run_training(training_dataset_file, validation_dataset_file, obj,
                 nb_epochs, out_checkpoints_folder, out_logs_folder,
                 validation_interval, use_uncertainty, save_weights_only=False):
    """
        Run the training for a given object id
    """
    cat_id = obj["category_id"]
    obj_id = obj["object_id"]
    print("========================================")
    print("  Start training object: %d (class %d)" % (obj_id, cat_id))
    print("========================================")
    name = "obj_%02d_%02d" % (cat_id, obj_id)
    dataset_train = EllipsesDataset(training_dataset_file, obj_id,
                                    crop_size=_cfg.CROP_SIZE,
                                    transforms=transforms.Compose([
                                        transforms.ColorJitter(*_cfg.TRAIN_COLOR_JITTER),
                                        transforms.ToTensor()]),
                                    random_shift=_cfg.TRAIN_RANDOM_SHIFT,
                                    random_rotation=_cfg.TRAIN_RANDOM_ROTATION,
                                    random_perspective=_cfg.TRAIN_RANDOM_PERSPECTIVE,
                                    random_blur=_cfg.TRAIN_RANDOM_BLUR,
                                    probability_do_augm=_cfg.TRAIN_PROBA_DO_AUGM)

    dataset_valid = EllipsesDataset(validation_dataset_file, obj_id, 
                                    crop_size=_cfg.CROP_SIZE,
                                    transforms=transforms.Compose([
                                        transforms.ColorJitter(*_cfg.VALID_COLOR_JITTER),
                                        transforms.ToTensor()]),
                                    random_shift=_cfg.VALID_RANDOM_SHIFT,
                                    random_rotation=_cfg.VALID_RANDOM_ROTATION,
                                    random_perspective=_cfg.VALID_RANDOM_PERSPECTIVE,
                                    random_blur=_cfg.VALID_RANDOM_BLUR)


    tb_logger = pl_loggers.TensorBoardLogger(os.path.join(out_logs_folder, name))

    training_loader = torch.utils.data.DataLoader(dataset_train, 
                                                  batch_size=_cfg.TRAIN_BATCH_SIZE,
                                                  shuffle=_cfg.TRAIN_BATCH_SHUFFLE,
                                                  num_workers=_cfg.TRAIN_LOADER_NUM_WORKERS,
                                                  drop_last=True)

    validation_loader = torch.utils.data.DataLoader(dataset_valid, 
                                                    batch_size=_cfg.VALID_BATCH_SIZE,
                                                    shuffle=_cfg.VALID_BATCH_SHUFFLE,
                                                    num_workers=_cfg.VALID_LOADER_NUM_WORKERS)
    # Model
    if use_uncertainty:
        model = EllipsePredictor_uncertainty(lr=_cfg.LEARNING_RATE,
                                             crop_size=_cfg.CROP_SIZE,
                                             output_val_images=None)
    else:
        model = EllipsePredictor(lr=_cfg.LEARNING_RATE,
                                 crop_size=_cfg.CROP_SIZE,
                                 output_val_images=None)

    # Checkpoint callback
    checkpoint_callback = ModelCheckpoint(
        dirpath=os.path.join(out_checkpoints_folder, name),
        save_last=True,
        filename="best",
        save_top_k=1,
        verbose=True,
        monitor='val_ious',
        mode='max',
        prefix='ckpt',
        save_weights_only=save_weights_only
    )


    # Training
    ta = time.time()

    trainer = pl.Trainer(gpus=1, logger=tb_logger, max_epochs=nb_epochs,
                         check_val_every_n_epoch=validation_interval,
                         log_every_n_steps=50,
                         checkpoint_callback=checkpoint_callback)
    trainer.fit(model, training_loader, validation_loader)

    print("Finished in %.2fs" % (time.time()-ta))


def main(args):
    parser = argparse.ArgumentParser()
    parser.add_argument("scene", help="<Required> Input Scene file containing the objects (.json)")
    parser.add_argument("training", help="<Required> Input training dataset file (.json)")
    parser.add_argument("validation", help="<Required> Input validation dataset file (.json)")
    parser.add_argument("checkpoints", help="<Required> Output folder where to save checkpoints.")
    parser.add_argument("--logs", help="Output folder where to output training logs.",
                        default="logs")
    parser.add_argument("--nb_epochs", help="Number of training epochs for each object.",
                        default=300, type=int)
    parser.add_argument("--valid_interval", help="Interval of epochs between validations.",
                        default=10, type=int)
    parser.add_argument("--object_id", help="<Optional> Specify a single object id for training.",
                        default=None, type=int)
    parser.add_argument("--save_all_parameters", action="store_true", 
                        help="Save all the hyper-parameters in checkpoints, "
                             "not only weights (default is False).",
                        default=False)
    parser.add_argument("--use_uncertainty",  action="store_true", default=False,
                        help="<Optional> Enable usage of uncertainty (default is False).")

    args = parser.parse_args(args)


    scene_file = args.scene
    training_dataset_file = args.training
    validation_dataset_file = args.validation
    out_checkpoints_folder = args.checkpoints
    out_logs_folder = args.logs
    nb_epochs = args.nb_epochs
    validation_interval = args.valid_interval
    save_all_parameters = args.save_all_parameters
    save_weights_only = not save_all_parameters
    only_object_id = args.object_id
    use_uncertainty = args.use_uncertainty
    if use_uncertainty:
        print("Train model with uncertainty.")
    else:
        print("Train model without uncertainty.")

    scene = Scene_loader(scene_file)

    print("List of objects:")
    for obj in scene:
        print(" - obj_%02d_%02d" % (obj["category_id"], obj["object_id"]))
    print()

    if only_object_id is not None:
        print("Train only object", only_object_id)
        scene = scene[only_object_id:only_object_id+1]

    for obj in scene:
        run_training(training_dataset_file, validation_dataset_file, obj,
                     nb_epochs=nb_epochs,
                     out_checkpoints_folder=out_checkpoints_folder,
                     out_logs_folder=out_logs_folder,
                     validation_interval=validation_interval,
                     use_uncertainty=use_uncertainty,
                     save_weights_only=save_weights_only)



if __name__ == '__main__':
    import sys
    try:
        main(sys.argv[1:])
    except Exception as e:
        logging.exception(e)
        sys.exit(1)